#include <stdio.h>
#include <ilib.h>
#include <memory.h>
#include <sys/errno.h>
#include "config.h"
#include <pass.h>

#include "sys/profiler.h"
#include <pcap.h>

#include <sys/netio.h>

#include <sys/simulator.h>
#include <netinet/in.h>

#define PROFILE_DETECT
//#define HASH_PACKETS
#define MEASURE_PERFORMANCE
#ifdef MEASURE_PERFORMANCE
// Exported from application
uint64_t       pcap_start_cycle;
uint64_t       pcap_end_cycle;
#endif

// For hashing from file
extern uint          tile_count;

#ifdef PROFILE_NETIO
static unsigned int empty_requests = 0;  // Request for packet that are not fulfilled.
static uint64_t    cycles_waiting_for_packets = 0;
static unsigned int packets_processed = 0;
#endif

//
// Local
//
typedef enum { NO_DEVICE, XGBE0, XGBE1 } MyDevices;
static MyDevices my_device = NO_DEVICE;

//
// From NetIO implementation, only 32 tiles can register for anyone queue.
//
#define MAX_TILES_PER_QUEUE 32

//--------------------------------------------------------------------------------
//
// from libpcap-8.3/savefile.c
//

#define HAVE_SNPRINTF  1
#define HAVE_VSNPRINTF 1
#include "pcap-int.h"

#define TCPDUMP_MAGIC 0xa1b2c3d4
#define PATCHED_TCPDUMP_MAGIC 0xa1b2cd34

/*
 * We use the "receiver-makes-right" approach to byte order,
 * because time is at a premium when we are writing the file.
 * In other words, the pcap_file_header and pcap_pkthdr,
 * records are written in host byte order.
 * Note that the bytes of packet data are written out in the order in
 * which they were received, so multi-byte fields in packets are not
 * written in host byte order, they're written in whatever order the
 * sending machine put them in.
 *
 * ntoh[ls] aren't sufficient because we might need to swap on a big-endian
 * machine (if the file was written in little-end order).
 */
#define	SWAPLONG(y)							\
  ((((y)&0xff)<<24) | (((y)&0xff00)<<8) | (((y)&0xff0000)>>8) | (((y)>>24)&0xff))
#define	SWAPSHORT(y)					\
  ( (((y)&0xff)<<8) | ((u_short)((y)&0xff00)>>8) )

static void
swap_hdr(struct pcap_file_header *hp)
{
  hp->version_major = SWAPSHORT(hp->version_major);
  hp->version_minor = SWAPSHORT(hp->version_minor);
  hp->thiszone = SWAPLONG(hp->thiszone);
  hp->sigfigs = SWAPLONG(hp->sigfigs);
  hp->snaplen = SWAPLONG(hp->snaplen);
  hp->linktype = SWAPLONG(hp->linktype);
}

#define SFERR_TRUNC		1
#define SFERR_BADVERSION	2
#define SFERR_BADF		3
#define SFERR_EOF		4 /* not really an error, just a status */

/*
 * We don't write DLT_* values to the capture file header, because
 * they're not the same on all platforms.
 *
 * Unfortunately, the various flavors of BSD have not always used the same
 * numerical values for the same data types, and various patches to
 * libpcap for non-BSD OSes have added their own DLT_* codes for link
 * layer encapsulation types seen on those OSes, and those codes have had,
 * in some cases, values that were also used, on other platforms, for other
 * link layer encapsulation types.
 *
 * This means that capture files of a type whose numerical DLT_* code
 * means different things on different BSDs, or with different versions
 * of libpcap, can't always be read on systems other than those like
 * the one running on the machine on which the capture was made.
 *
 * Instead, we define here a set of LINKTYPE_* codes, and map DLT_* codes
 * to LINKTYPE_* codes when writing a savefile header, and map LINKTYPE_*
 * codes to DLT_* codes when reading a savefile header.
 *
 * For those DLT_* codes that have, as far as we know, the same values on
 * all platforms (DLT_NULL through DLT_FDDI), we define LINKTYPE_xxx as
 * DLT_xxx; that way, captures of those types can still be read by
 * versions of libpcap that map LINKTYPE_* values to DLT_* values, and
 * captures of those types written by versions of libpcap that map DLT_
 * values to LINKTYPE_ values can still be read by older versions
 * of libpcap.
 *
 * The other LINKTYPE_* codes are given values starting at 100, in the
 * hopes that no DLT_* code will be given one of those values.
 *
 * In order to ensure that a given LINKTYPE_* code's value will refer to
 * the same encapsulation type on all platforms, you should not allocate
 * a new LINKTYPE_* value without consulting "tcpdump-workers@tcpdump.org".
 * The tcpdump developers will allocate a value for you, and will not
 * subsequently allocate it to anybody else; that value will be added to
 * the "pcap.h" in the tcpdump.org CVS repository, so that a future
 * libpcap release will include it.
 *
 * You should, if possible, also contribute patches to libpcap and tcpdump
 * to handle the new encapsulation type, so that they can also be checked
 * into the tcpdump.org CVS repository and so that they will appear in
 * future libpcap and tcpdump releases.
 *
 * Do *NOT* assume that any values after the largest value in this file
 * are available; you might not have the most up-to-date version of this
 * file, and new values after that one might have been assigned.  Also,
 * do *NOT* use any values below 100 - those might already have been
 * taken by one (or more!) organizations.
 */
#define LINKTYPE_NULL		DLT_NULL
#define LINKTYPE_ETHERNET	DLT_EN10MB	/* also for 100Mb and up */
#define LINKTYPE_EXP_ETHERNET	DLT_EN3MB	/* 3Mb experimental Ethernet */
#define LINKTYPE_AX25		DLT_AX25
#define LINKTYPE_PRONET		DLT_PRONET
#define LINKTYPE_CHAOS		DLT_CHAOS
#define LINKTYPE_TOKEN_RING	DLT_IEEE802	/* DLT_IEEE802 is used for Token Ring */
#define LINKTYPE_ARCNET		DLT_ARCNET	/* BSD-style headers */
#define LINKTYPE_SLIP		DLT_SLIP
#define LINKTYPE_PPP		DLT_PPP
#define LINKTYPE_FDDI		DLT_FDDI

/*
 * LINKTYPE_PPP is for use when there might, or might not, be an RFC 1662
 * PPP in HDLC-like framing header (with 0xff 0x03 before the PPP protocol
 * field) at the beginning of the packet.
 *
 * This is for use when there is always such a header; the address field
 * might be 0xff, for regular PPP, or it might be an address field for Cisco
 * point-to-point with HDLC framing as per section 4.3.1 of RFC 1547 ("Cisco
 * HDLC").  This is, for example, what you get with NetBSD's DLT_PPP_SERIAL.
 *
 * We give it the same value as NetBSD's DLT_PPP_SERIAL, in the hopes that
 * nobody else will choose a DLT_ value of 50, and so that DLT_PPP_SERIAL
 * captures will be written out with a link type that NetBSD's tcpdump
 * can read.
 */
#define LINKTYPE_PPP_HDLC	50		/* PPP in HDLC-like framing */

#define LINKTYPE_PPP_ETHER	51		/* NetBSD PPP-over-Ethernet */

/*
 * This isn't supported in libpcap 0.8[.x], but is supported in the
 * current CVS version; we include it here to note that it's not available
 * for anybody else to use.
 */
#define LINKTYPE_SYMANTEC_FIREWALL 99		/* Symantec Enterprise Firewall */

#define LINKTYPE_ATM_RFC1483	100		/* LLC/SNAP-encapsulated ATM */
#define LINKTYPE_RAW		101		/* raw IP */
#define LINKTYPE_SLIP_BSDOS	102		/* BSD/OS SLIP BPF header */
#define LINKTYPE_PPP_BSDOS	103		/* BSD/OS PPP BPF header */
#define LINKTYPE_C_HDLC		104		/* Cisco HDLC */
#define LINKTYPE_IEEE802_11	105		/* IEEE 802.11 (wireless) */
#define LINKTYPE_ATM_CLIP	106		/* Linux Classical IP over ATM */
#define LINKTYPE_FRELAY		107		/* Frame Relay */
#define LINKTYPE_LOOP		108		/* OpenBSD loopback */
#define LINKTYPE_ENC		109		/* OpenBSD IPSEC enc */

/*
 * These three types are reserved for future use.
 */
#define LINKTYPE_LANE8023	110		/* ATM LANE + 802.3 */
#define LINKTYPE_HIPPI		111		/* NetBSD HIPPI */
#define LINKTYPE_HDLC		112		/* NetBSD HDLC framing */

#define LINKTYPE_LINUX_SLL	113		/* Linux cooked socket capture */
#define LINKTYPE_LTALK		114		/* Apple LocalTalk hardware */
#define LINKTYPE_ECONET		115		/* Acorn Econet */

/*
 * Reserved for use with OpenBSD ipfilter.
 */
#define LINKTYPE_IPFILTER	116

#define LINKTYPE_PFLOG		117		/* OpenBSD DLT_PFLOG */
#define LINKTYPE_CISCO_IOS	118		/* For Cisco-internal use */
#define LINKTYPE_PRISM_HEADER	119		/* 802.11+Prism II monitor mode */
#define LINKTYPE_AIRONET_HEADER	120		/* FreeBSD Aironet driver stuff */

/*
 * Reserved for Siemens HiPath HDLC.
 */
#define LINKTYPE_HHDLC		121

#define LINKTYPE_IP_OVER_FC	122		/* RFC 2625 IP-over-Fibre Channel */
#define LINKTYPE_SUNATM		123		/* Solaris+SunATM */

/*
 * Reserved as per request from Kent Dahlgren <kent@praesum.com>
 * for private use.
 */
#define LINKTYPE_RIO		124		/* RapidIO */
#define LINKTYPE_PCI_EXP	125		/* PCI Express */
#define LINKTYPE_AURORA		126		/* Xilinx Aurora link layer */

#define LINKTYPE_IEEE802_11_RADIO 127		/* 802.11 plus BSD radio header */

/*
 * Reserved for the TZSP encapsulation, as per request from
 * Chris Waters <chris.waters@networkchemistry.com>
 * TZSP is a generic encapsulation for any other link type,
 * which includes a means to include meta-information
 * with the packet, e.g. signal strength and channel
 * for 802.11 packets.
 */
#define LINKTYPE_TZSP		128		/* Tazmen Sniffer Protocol */

#define LINKTYPE_ARCNET_LINUX	129		/* Linux-style headers */

/*
 * Juniper-private data link types, as per request from
 * Hannes Gredler <hannes@juniper.net>.  The corresponding
 * DLT_s are used for passing on chassis-internal
 * metainformation such as QOS profiles, etc..
 */
#define LINKTYPE_JUNIPER_MLPPP  130
#define LINKTYPE_JUNIPER_MLFR   131
#define LINKTYPE_JUNIPER_ES     132
#define LINKTYPE_JUNIPER_GGSN   133
#define LINKTYPE_JUNIPER_MFR    134
#define LINKTYPE_JUNIPER_ATM2   135
#define LINKTYPE_JUNIPER_SERVICES 136
#define LINKTYPE_JUNIPER_ATM1   137

#define LINKTYPE_APPLE_IP_OVER_IEEE1394 138	/* Apple IP-over-IEEE 1394 cooked header */

#define LINKTYPE_RAWSS7         139             /* see rawss7.h for */
#define LINKTYPE_RAWSS7_MTP2    140	        /* information  on these */
#define LINKTYPE_RAWSS7_MTP3    141             /* definitions */
#define LINKTYPE_RAWSS7_SCCP    142

/*
 * This isn't supported in libpcap 0.8[.x], but is supported in the
 * current CVS version; we include it here to note that it's not available
 * for anybody else to use.
 */
#define LINKTYPE_DOCSIS		143		/* DOCSIS MAC frames */

#define LINKTYPE_LINUX_IRDA	144		/* Linux-IrDA */

/*
 * Reserved for IBM SP switch and IBM Next Federation switch.
 */
#define LINKTYPE_IBM_SP		145
#define LINKTYPE_IBM_SN		146

/*
 * Reserved for private use.  If you have some link-layer header type
 * that you want to use within your organization, with the capture files
 * using that link-layer header type not ever be sent outside your
 * organization, you can use these values.
 *
 * No libpcap release will use these for any purpose, nor will any
 * tcpdump release use them, either.
 *
 * Do *NOT* use these in capture files that you expect anybody not using
 * your private versions of capture-file-reading tools to read; in
 * particular, do *NOT* use them in products, otherwise you may find that
 * people won't be able to use tcpdump, or snort, or Ethereal, or... to
 * read capture files from your firewall/intrusion detection/traffic
 * monitoring/etc. appliance, or whatever product uses that LINKTYPE_ value,
 * and you may also find that the developers of those applications will
 * not accept patches to let them read those files.
 *
 * Also, do not use them if somebody might send you a capture using them
 * for *their* private type and tools using them for *your* private type
 * would have to read them.
 *
 * Instead, in those cases, ask "tcpdump-workers@tcpdump.org" for a new DLT_
 * and LINKTYPE_ value, as per the comment in pcap-bpf.h, and use the type
 * you're given.
 */
#define LINKTYPE_USER0		147
#define LINKTYPE_USER1		148
#define LINKTYPE_USER2		149
#define LINKTYPE_USER3		150
#define LINKTYPE_USER4		151
#define LINKTYPE_USER5		152
#define LINKTYPE_USER6		153
#define LINKTYPE_USER7		154
#define LINKTYPE_USER8		155
#define LINKTYPE_USER9		156
#define LINKTYPE_USER10		157
#define LINKTYPE_USER11		158
#define LINKTYPE_USER12		159
#define LINKTYPE_USER13		160
#define LINKTYPE_USER14		161
#define LINKTYPE_USER15		162

/*
 * For future use with 802.11 captures - defined by AbsoluteValue
 * Systems to store a number of bits of link-layer information
 * including radio information:
 *
 *	http://www.shaftnet.org/~pizza/software/capturefrm.txt
 *
 * but could and arguably should also be used by non-AVS Linux
 * 802.11 drivers; that may happen in the future.
 */
#define LINKTYPE_IEEE802_11_RADIO_AVS 163	/* 802.11 plus AVS radio header */

/*
 * Juniper-private data link type, as per request from
 * Hannes Gredler <hannes@juniper.net>.  The corresponding
 * DLT_s are used for passing on chassis-internal
 * metainformation such as QOS profiles, etc..
 */
#define LINKTYPE_JUNIPER_MONITOR 164

static struct linktype_map {
  int	dlt;
  int	linktype;
} map[] = {
  /*
   * These DLT_* codes have LINKTYPE_* codes with values identical
   * to the values of the corresponding DLT_* code.
   */
  { DLT_NULL,		LINKTYPE_NULL },
  { DLT_EN10MB,		LINKTYPE_ETHERNET },
  { DLT_EN3MB,		LINKTYPE_EXP_ETHERNET },
  { DLT_AX25,		LINKTYPE_AX25 },
  { DLT_PRONET,		LINKTYPE_PRONET },
  { DLT_CHAOS,		LINKTYPE_CHAOS },
  { DLT_IEEE802,		LINKTYPE_TOKEN_RING },
  { DLT_ARCNET,		LINKTYPE_ARCNET },
  { DLT_SLIP,		LINKTYPE_SLIP },
  { DLT_PPP,		LINKTYPE_PPP },
  { DLT_FDDI,	 	LINKTYPE_FDDI },

  /*
   * These DLT_* codes have different values on different
   * platforms; we map them to LINKTYPE_* codes that
   * have values that should never be equal to any DLT_*
   * code.
   */
#ifdef DLT_FR
  /* BSD/OS Frame Relay */
  { DLT_FR,		LINKTYPE_FRELAY },
#endif

  { DLT_SYMANTEC_FIREWALL, LINKTYPE_SYMANTEC_FIREWALL },
  { DLT_ATM_RFC1483, 	LINKTYPE_ATM_RFC1483 },
  { DLT_RAW,		LINKTYPE_RAW },
  { DLT_SLIP_BSDOS,	LINKTYPE_SLIP_BSDOS },
  { DLT_PPP_BSDOS,	LINKTYPE_PPP_BSDOS },

  /* BSD/OS Cisco HDLC */
  { DLT_C_HDLC,		LINKTYPE_C_HDLC },

  /*
   * These DLT_* codes are not on all platforms, but, so far,
   * there don't appear to be any platforms that define
   * other codes with those values; we map them to
   * different LINKTYPE_* values anyway, just in case.
   */

  /* Linux ATM Classical IP */
  { DLT_ATM_CLIP,		LINKTYPE_ATM_CLIP },

  /* NetBSD sync/async serial PPP (or Cisco HDLC) */
  { DLT_PPP_SERIAL,	LINKTYPE_PPP_HDLC },

  /* NetBSD PPP over Ethernet */
  { DLT_PPP_ETHER,	LINKTYPE_PPP_ETHER },

  /* IEEE 802.11 wireless */
  { DLT_IEEE802_11,	LINKTYPE_IEEE802_11 },

  /* Frame Relay */
  { DLT_FRELAY,		LINKTYPE_FRELAY },

  /* OpenBSD loopback */
  { DLT_LOOP,		LINKTYPE_LOOP },

  /* Linux cooked socket capture */
  { DLT_LINUX_SLL,	LINKTYPE_LINUX_SLL },

  /* Apple LocalTalk hardware */
  { DLT_LTALK,		LINKTYPE_LTALK },

  /* Acorn Econet */
  { DLT_ECONET,		LINKTYPE_ECONET },

  /* OpenBSD DLT_PFLOG */
  { DLT_PFLOG,		LINKTYPE_PFLOG },

  /* For Cisco-internal use */
  { DLT_CISCO_IOS,	LINKTYPE_CISCO_IOS },

  /* Prism II monitor-mode header plus 802.11 header */
  { DLT_PRISM_HEADER,	LINKTYPE_PRISM_HEADER },

  /* FreeBSD Aironet driver stuff */
  { DLT_AIRONET_HEADER,	LINKTYPE_AIRONET_HEADER },

  /* Siemens HiPath HDLC */
  { DLT_HHDLC,		LINKTYPE_HHDLC },

  /* RFC 2625 IP-over-Fibre Channel */
  { DLT_IP_OVER_FC,	LINKTYPE_IP_OVER_FC },

  /* Solaris+SunATM */
  { DLT_SUNATM,		LINKTYPE_SUNATM },

  /* RapidIO */
  { DLT_RIO,		LINKTYPE_RIO },

  /* PCI Express */
  { DLT_PCI_EXP,		LINKTYPE_PCI_EXP },

  /* Xilinx Aurora link layer */
  { DLT_AURORA,		LINKTYPE_AURORA },

  /* 802.11 plus BSD radio header */
  { DLT_IEEE802_11_RADIO,	LINKTYPE_IEEE802_11_RADIO },

  /* Tazmen Sniffer Protocol */
  { DLT_TZSP,		LINKTYPE_TZSP },

  /* Arcnet with Linux-style link-layer headers */
  { DLT_ARCNET_LINUX,	LINKTYPE_ARCNET_LINUX },

  /* Juniper-internal chassis encapsulation */
  { DLT_JUNIPER_MLPPP,    LINKTYPE_JUNIPER_MLPPP },
  { DLT_JUNIPER_MLFR,     LINKTYPE_JUNIPER_MLFR },
  { DLT_JUNIPER_ES,       LINKTYPE_JUNIPER_ES },
  { DLT_JUNIPER_GGSN,     LINKTYPE_JUNIPER_GGSN },
  { DLT_JUNIPER_MFR,      LINKTYPE_JUNIPER_MFR },
  { DLT_JUNIPER_ATM2,     LINKTYPE_JUNIPER_ATM2 },
  { DLT_JUNIPER_SERVICES, LINKTYPE_JUNIPER_SERVICES },
  { DLT_JUNIPER_ATM1,     LINKTYPE_JUNIPER_ATM1 },

  /* Apple IP-over-IEEE 1394 cooked header */
  { DLT_APPLE_IP_OVER_IEEE1394, LINKTYPE_APPLE_IP_OVER_IEEE1394 },

  /* DOCSIS MAC frames */
  { DLT_DOCSIS,		LINKTYPE_DOCSIS },

  /* IrDA IrLAP packets + Linux-cooked header */
  { DLT_LINUX_IRDA,	LINKTYPE_LINUX_IRDA },

  /* IBM SP and Next Federation switches */
  { DLT_IBM_SP,		LINKTYPE_IBM_SP },
  { DLT_IBM_SN,		LINKTYPE_IBM_SN },

  /* 802.11 plus AVS radio header */
  { DLT_IEEE802_11_RADIO_AVS, LINKTYPE_IEEE802_11_RADIO_AVS },

  /*
   * Any platform that defines additional DLT_* codes should:
   *
   *	request a LINKTYPE_* code and value from tcpdump.org,
   *	as per the above;
   *
   *	add, in their version of libpcap, an entry to map
   *	those DLT_* codes to the corresponding LINKTYPE_*
   *	code;
   *
   *	redefine, in their "net/bpf.h", any DLT_* values
   *	that collide with the values used by their additional
   *	DLT_* codes, to remove those collisions (but without
   *	making them collide with any of the LINKTYPE_*
   *	values equal to 50 or above; they should also avoid
   *	defining DLT_* values that collide with those
   *	LINKTYPE_* values, either).
   */

  /* Juniper-internal chassis encapsulation */
  { DLT_JUNIPER_MONITOR,     LINKTYPE_JUNIPER_MONITOR },

  { -1,			-1 }
};

static int
dlt_to_linktype(int dlt)
{
  int i;

  for (i = 0; map[i].dlt != -1; i++) {
    if (map[i].dlt == dlt)
      return (map[i].linktype);
  }

  /*
   * If we don't have a mapping for this DLT_ code, return an
   * error; that means that the table above needs to have an
   * entry added.
   */
  return (-1);
}

static int
linktype_to_dlt(int linktype)
{
  int i;

  for (i = 0; map[i].linktype != -1; i++) {
    if (map[i].linktype == linktype)
      return (map[i].dlt);
  }

  /*
   * If we don't have an entry for this link type, return
   * the link type value; it may be a DLT_ value from an
   * older version of libpcap.
   */
  return linktype;
}

static int
sf_write_header(FILE *fp, int linktype, int thiszone, int snaplen)
{
  struct pcap_file_header hdr;

  hdr.magic = TCPDUMP_MAGIC;
  hdr.version_major = PCAP_VERSION_MAJOR;
  hdr.version_minor = PCAP_VERSION_MINOR;

  hdr.thiszone = thiszone;
  hdr.snaplen = snaplen;
  hdr.sigfigs = 0;
  hdr.linktype = linktype;

  if (fwrite((char *)&hdr, sizeof(hdr), 1, fp) != 1)
    return (-1);

  return (0);
}

/*
 * Output a packet to the initialized dump file.
 */
void
pcap_dump(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
  register FILE *f;
  struct pcap_sf_pkthdr sf_hdr;

  f = (FILE *)user;
  sf_hdr.ts.tv_sec  = h->ts.tv_sec;
  sf_hdr.ts.tv_usec = h->ts.tv_usec;
  sf_hdr.caplen     = h->caplen;
  sf_hdr.len        = h->len;
  /* XXX we should check the return status */
  (void)fwrite(&sf_hdr, sizeof(sf_hdr), 1, f);
  (void)fwrite((char *)sp, h->caplen, 1, f);
}

/*
 * Initialize so that sf_write() will output to the file named 'fname'.
 */
pcap_dumper_t *
pcap_dump_open(pcap_t *p, const char *fname)
{
  FILE *f;
  int linktype;

  linktype = dlt_to_linktype(p->linktype);
  if (linktype == -1) {
    snprintf(p->errbuf, PCAP_ERRBUF_SIZE,
	     "%s: link-layer type %d isn't supported in savefiles",
	     fname, linktype);
    return (NULL);
  }

  if (fname[0] == '-' && fname[1] == '\0') {
    f = stdout;
#ifdef WIN32
    _setmode(_fileno(f), _O_BINARY);
#endif
  } else {
#ifndef WIN32
    f = fopen(fname, "w");
#else
    f = fopen(fname, "wb");
    setbuf(f, NULL);	/* XXX - why? */
#endif
    if (f == NULL) {
      snprintf(p->errbuf, PCAP_ERRBUF_SIZE, "%s: %s",
	       fname, pcap_strerror(errno));
      return (NULL);
    }
  }
  (void)sf_write_header(f, linktype, p->tzoff, p->snapshot);
  return ((pcap_dumper_t *)f);
}

//--------------------------------------------------------------------------------

static   FILE * input_fp = 0;

#define MAX_RECEIVE_PACKETS           2048
#define MAX_FILE_PACKETS              250000

static   netio_queue_t ipp_queue;
static   int ipp_initialized = 0;

#define RR_QUEUES   4

static void configure_ipp(netio_queue_t* queue)
{
  netio_error_t rc;
  netio_input_config_t config;
  memset(&config, 0, sizeof(config));

  int group_size = ilib_group_size(ILIB_GROUP_SIBLINGS);
  int group_rank = ilib_group_rank(ILIB_GROUP_SIBLINGS);

  config.num_receive_packets = MAX_RECEIVE_PACKETS;
  if (my_device == XGBE1)
    {// XAUI 1
      config.interface = "xgbe/1";
    }
  else
    {// XAUI 0
      config.interface = "xgbe/0";
    }

#ifdef ROUND_ROBIN
  // Can't have more than 32 tiles in one Round Robin group, so
  // For more than 32 tiles, divide into multiple groups (RR_QUEUES).
  config.queue_id = (group_size > MAX_TILES_PER_QUEUE ? group_rank % RR_QUEUES : 0);
#else
  config.queue_id = group_rank;
#endif

  config.num_receive_packets = MAX_RECEIVE_PACKETS;
  config.num_send_buffers_small_total = 0;
  config.num_send_buffers_large_total = 0;
  config.num_send_buffers_jumbo_total = 0;
  config.num_send_buffers_small_prealloc = 0;
  config.num_send_buffers_large_prealloc = 0;
  config.num_send_buffers_jumbo_prealloc = 0;


  //
  // Only allow one NetIO input register at a time.
  //
  for (int order = 0; order < group_size; order++)
    {
      if (group_rank == order)
	{
	  if ((rc=netio_input_register(&config, queue)) != NETIO_NO_ERROR)
	    {
	      ilib_die("%d in 'netio_input_register()': %s\n",
		       rc, netio_strerror(rc));
	    }
	}
      ilib_msg_barrier(ILIB_GROUP_SIBLINGS);
    }
  ilib_msg_barrier(ILIB_GROUP_SIBLINGS);

  int total_buckets = 512;
  int total_queues = group_size;

  if (group_rank == 0)
    {
      //
      // Configure buckets to hash sessions across the number of Snort tiles.
      //

      //
      // Must configure VLANs to cause packets to be round-robin distributed
      // Map all VLANs the same way. The VLAN tag is 12 bits, so 4096 different VLAN
      // tags are possible.
      //
      for (int vv = 0; vv < 4096; vv++)
	{
	  netio_vlan_t vlan;
	  memset(&vlan, 0, sizeof(vlan));

	  vlan.bits.__bucket_base = 0;
#ifdef ROUND_ROBIN
	  if (group_size <= MAX_TILES_PER_QUEUE)
	    {
	      // All tiles Round-Robin from one queue
	      total_buckets = 1;
	      total_queues = 1;
	    }
	  else
	    {
	      // Half the tiles round-robin from queue 0, the others from queue 1 (odd/even)
	      total_buckets = RR_QUEUES;
	      total_queues = RR_QUEUES;
	    }
#endif

	  vlan.bits.__bucket_mask = total_buckets-1;  // 512 buckets

	  if (total_queues > 1)
	    {
	      vlan.bits.__balance_on_l4 = 1;  // Hash ports?
	      vlan.bits.__balance_on_l3 = 1;  // Hash based on IP addresses?
	      vlan.bits.__balance_on_l2 = 1;  // Hash based on Ethernet Mac address
	    }
	  else
	    {
	      vlan.bits.__balance_on_l4 = 0;  // Hash ports?
	      vlan.bits.__balance_on_l3 = 0;  // Hash based on IP addresses?
	      vlan.bits.__balance_on_l2 = 0;  // Hash based on Ethernet Mac address
	    }

	  //
	  // Configure one VLAN starting at VLAN 0
	  //
	  if ((rc = netio_input_vlan_configure(queue, 0, &vlan, 1)) != NETIO_NO_ERROR)
	    {
	      printf("Error: Error %d in 'netio_input_vlan_configure()': %s\n", rc, netio_strerror(rc));
	    }
	}

      //
      // Create mapping from hash buckets to queue numbers, where each tile has its own queu
      //
      netio_bucket_t bucket_map[512];
      int b = 0;
      for (int i = 0; i < total_buckets; i++)
	{
	  bucket_map[i] = b;
	  b++;
	  if (b >= total_queues)
	    b = 0;
	}
      if ((rc=netio_input_bucket_configure(queue, 0, bucket_map, total_buckets)) != NETIO_NO_ERROR)
	{
	  printf("Error: Error %d in 'netio_input_bucket_configure()': %s\n", rc, netio_strerror(rc));
	}


      if ((rc=netio_input_initialize(queue)) != NETIO_NO_ERROR)
	{
	  printf("Error: Error %d in 'netio_input_initialize()': %s\n", rc, netio_strerror(rc));
	}

      //printf("IPP initialized from (%d, %d), buckets=%d, queues=%d\n", spr_x_coordinate(), spr_y_coordinate(), total_buckets, total_queues);
    }
}

static void pcap_close_dead(pcap_t *p);


char	*pcap_lookupdev(char *x) { return "eth0"; }

int
pcap_lookupnet(device, netp, maskp, errbuf)
     register const char *device;
     register bpf_u_int32 *netp, *maskp;
     register char *errbuf;
{
  /*
   * The pseudo-device "any" listens on all interfaces and therefore
   * has the network address and -mask "0.0.0.0" therefore catching
   * all traffic. Using NULL for the interface is the same as "any".
   */
  if (!device || strcmp(device, "any") == 0)
    {
      *netp = *maskp = 0;
      return 0;
    }
  return -1;
}

static void
pcap_close_tilera_ipp(pcap_t *p)
{
  //printf("Closing live PCAP. rank %d of %d\n",ilib_group_rank(ILIB_GROUP_SIBLINGS), ilib_group_size(ILIB_GROUP_SIBLINGS) );
  /* Tell IPP to quit, but only after all the tiles are done, and then only once */

  ilib_msg_barrier(ILIB_GROUP_SIBLINGS);

#if 1
  if (ilib_group_rank(ILIB_GROUP_SIBLINGS) == 0)
    {
      // printf("Shutting down IPP\n");
      fflush(stdout);

      // __netio_input_uninitialize(&ipp_queue);
    }
#endif
}

static int
pcap_stats_tilera_ipp(pcap_t *p, struct pcap_stat *ps)
{
  return (-1);
}

pcap_t *
pcap_open_offline(const char *fname, char *errbuf)
{
  register pcap_t *p;
  register FILE *fp;
  struct pcap_file_header hdr;
  bpf_u_int32 magic;
  int linklen;

  p = (pcap_t *)malloc_shared(sizeof(*p));
  if (p == NULL) {
    strlcpy(errbuf, "out of swap", PCAP_ERRBUF_SIZE);
    return (NULL);
  }

  memset((char *)p, 0, sizeof(*p));

  if (fname[0] == '-' && fname[1] == '\0')
    fp = stdin;
  else {
#ifndef WIN32
    fp = fopen(fname, "r");
#else
    fp = fopen(fname, "rb");
#endif
    if (fp == NULL) {
      snprintf(errbuf, PCAP_ERRBUF_SIZE, "%s: %s", fname,
	       pcap_strerror(errno));
      goto bad;
    }
  }
  if (fread((char *)&hdr, sizeof(hdr), 1, fp) != 1) {
    snprintf(errbuf, PCAP_ERRBUF_SIZE, "fread: %s",
	     pcap_strerror(errno));
    goto bad;
  }
  magic = hdr.magic;
  if (magic != TCPDUMP_MAGIC && magic != PATCHED_TCPDUMP_MAGIC) {
    magic = SWAPLONG(magic);
    if (magic != TCPDUMP_MAGIC && magic != PATCHED_TCPDUMP_MAGIC) {
      snprintf(errbuf, PCAP_ERRBUF_SIZE,
	       "bad dump file format");
      goto bad;
    }
    p->sf.swapped = 1;
    swap_hdr(&hdr);
  }
  if (magic == PATCHED_TCPDUMP_MAGIC) {
    /*
     * XXX - the patch that's in some versions of libpcap
     * changes the packet header but not the magic number;
     * we'd have to use some hacks^H^H^H^H^Hheuristics to
     * detect that.
     */
    p->sf.hdrsize = sizeof(struct pcap_sf_patched_pkthdr);
  } else
    p->sf.hdrsize = sizeof(struct pcap_sf_pkthdr);
  if (hdr.version_major < PCAP_VERSION_MAJOR) {
    snprintf(errbuf, PCAP_ERRBUF_SIZE, "archaic file format");
    goto bad;
  }
  p->tzoff = hdr.thiszone;
  p->snapshot = hdr.snaplen;
  p->linktype = linktype_to_dlt(hdr.linktype);
  p->sf.rfile = fp;
#ifndef WIN32
  p->bufsize = hdr.snaplen;
#else
  /* Allocate the space for pcap_pkthdr as well. It will be used by pcap_read_ex */
  p->bufsize = hdr.snaplen+sizeof(struct pcap_pkthdr);
#endif

  /* Align link header as required for proper data alignment */
  /* XXX should handle all types */
  switch (p->linktype) {

  case DLT_EN10MB:
    linklen = 14;
    break;

  case DLT_FDDI:
    linklen = 13 + 8;	/* fddi_header + llc */
    break;

  case DLT_NULL:
  default:
    linklen = 0;
    break;
  }

  if (p->bufsize < 0)
    p->bufsize = BPF_MAXBUFSIZE;
  p->sf.base = (u_char *)malloc_shared(p->bufsize + BPF_ALIGNMENT);
  if (p->sf.base == NULL) {
    strlcpy(errbuf, "out of swap", PCAP_ERRBUF_SIZE);
    goto bad;
  }
  p->buffer = p->sf.base + BPF_ALIGNMENT - (linklen % BPF_ALIGNMENT);
  p->sf.version_major = hdr.version_major;
  p->sf.version_minor = hdr.version_minor;
#ifdef PCAP_FDDIPAD
  /* XXX padding only needed for kernel fcode */
  pcap_fddipad = 0;
#endif

  /*
   * We interchanged the caplen and len fields at version 2.3,
   * in order to match the bpf header layout.  But unfortunately
   * some files were written with version 2.3 in their headers
   * but without the interchanged fields.
   *
   * In addition, DG/UX tcpdump writes out files with a version
   * number of 543.0, and with the caplen and len fields in the
   * pre-2.3 order.
   */
  switch (hdr.version_major) {

  case 2:
    if (hdr.version_minor < 3)
      p->sf.lengths_swapped = SWAPPED;
    else if (hdr.version_minor == 3)
      p->sf.lengths_swapped = MAYBE_SWAPPED;
    else
      p->sf.lengths_swapped = NOT_SWAPPED;
    break;

  case 543:
    p->sf.lengths_swapped = SWAPPED;
    break;

  default:
    p->sf.lengths_swapped = NOT_SWAPPED;
    break;
  }

#ifndef WIN32
  /*
   * You can do "select()" and "poll()" on plain files on most
   * platforms, and should be able to do so on pipes.
   *
   * You can't do "select()" on anything other than sockets in
   * Windows, so, on Win32 systems, we don't have "selectable_fd".
   */
  p->selectable_fd = fileno(fp);
#endif

  //p->read_op = pcap_offline_read;
  //p->setfilter_op = install_bpf_program;
  p->set_datalink_op = NULL;	/* we don't support munging link-layer headers */
  //p->stats_op = sf_stats;
  p->close_op = pcap_close_dead;

  input_fp = fp;
  return (p);
 bad:
  if(fp)
    fclose(fp);
  free(p);
  input_fp = NULL;
  return (NULL);
}

void
pcap_dump_close(pcap_dumper_t *p)
{
  (void)fclose((FILE *)p);
  input_fp = NULL;
}


/*
 *  Get a handle for a live capture from the given device. You can
 *  pass NULL as device to get all packages (without link level
 *  information of course). If you pass 1 as promisc the interface
 *  will be set to promiscous mode (XXX: I think this usage should
 *  be deprecated and functions be added to select that later allow
 *  modification of that values -- Torsten).
 *
 *  See also pcap(3).
 */
#define LINKTYPE_ETHERNET	DLT_EN10MB	/* also for 100Mb and up */

typedef struct {
  unsigned int   total_packets;
  ilibMutex      lock;
} TotalPacketsMutex;

pcap_t *
pcap_open_live(const char *device, int snaplen, int promisc, int to_ms,
               char *ebuf)
{
  pcap_t	*handle;
  int debug = 0;

  if(debug) printf("IPP PCAP open live (%d, %d) [rank=%d]\n",
		   spr_x_coordinate(), spr_y_coordinate(),
		   ilib_group_rank(ILIB_GROUP_SIBLINGS));

  if (strcmp(device, "xgbe1") == 0)
    {
      my_device = XGBE1;
      if(debug) printf("Configuring IPP for xgbe1\n");
    }
  else if (strcmp(device, "xgbe0") == 0)

    {
      my_device = XGBE0;
      if(debug) printf("Configuring IPP for xgbe0\n");
    }
  else
    {
      my_device = XGBE0;
      if(debug) printf("Warning: Unknown device - assuming xgbe0\n");
    }

  //
  // Create a shared data stucture for terminating simulation
  //
  if (ilib_group_rank(ILIB_GROUP_SIBLINGS) == 0)
    { // First Snort tiles does the allocation and initilization

      ((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex) = (TotalPacketsMutex * )malloc_shared(sizeof(TotalPacketsMutex));
      ((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex)->total_packets = 0;
      ilib_mutex_init(&	((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex)->lock);
    }
  ilibStatus status;
  ilib_msg_broadcast(ILIB_GROUP_SIBLINGS, 0, 
		     & ((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex), 
		     sizeof(((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex)), 
		     &status);

  /* Allocate a handle for this session. */
  if(debug) printf("[1] total_packets_mutex=%p [rank=%d]\n", 
		   ((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex),
		   ilib_group_rank(ILIB_GROUP_SIBLINGS));

  handle = malloc_shared(sizeof(pcap_t));
  if (handle == NULL) {
    fprintf(stderr, "malloc: %s", pcap_strerror(errno));
    return NULL;
  }
  memset(handle, 0, sizeof(pcap_t));
  handle->snapshot	= snaplen;
  handle->md.timeout	= to_ms;
  handle->md.stat.ps_recv = 0;
  handle->md.stat.ps_drop = 0;

  handle->linktype      = LINKTYPE_ETHERNET;
  handle->close_op = pcap_close_tilera_ipp;
  handle->stats_op = pcap_stats_tilera_ipp;
  //handle->read_op = pcap_read_tilera_ipp;


  return handle;
}

static int
pcap_stats_dead(pcap_t *p, struct pcap_stat *ps)
{
  snprintf(p->errbuf, PCAP_ERRBUF_SIZE,
	   "Statistics aren't available from a pcap_open_dead pcap_t");
  return (-1);
}

static void
pcap_close_dead(pcap_t *p)
{
  /* Nothing to do. */
}

pcap_t *
pcap_open_dead(int linktype, int snaplen)
{
  pcap_t *p;

  p = calloc(sizeof(*p), 1);
  if (p == NULL)
    return NULL;
  p->snapshot = snaplen;
  p->linktype = linktype;
  p->stats_op = pcap_stats_dead;
  p->close_op = pcap_close_dead;
  return p;
}




void
pcap_close(pcap_t *p)
{
  // assert (p->close_op == pcap_close_tilera_ipp || p->close_op == pcap_close_dead);
  p->close_op(p);
  if (p->dlt_list != NULL)
    free(p->dlt_list);
  pcap_freecode(&p->fcode);
  if (p)
    free(p);
}


//int	pcap_loop(pcap_t *a, int b, pcap_handler c, u_char *d)  { return 0; }

struct singleton {
  struct pcap_pkthdr *hdr;
  const u_char *pkt;
};

int
pcap_dispatch(pcap_t *p, int cnt, pcap_handler callback, u_char *user)
{

  return p->read_op(p, cnt, callback, user);
}

static inline void
invalidate_packet(unsigned char* data, unsigned int bytes_in_packet)
{
  int i;

  for (i = 0; i < bytes_in_packet; i += 16)
    __insn_finv(data + i);

  __insn_finv(data + bytes_in_packet - 1);

  __insn_mf();
}

void
invalidate_and_prefetch(netio_pkt_t* packet)
{
  netio_pkt_metadata_t* metadata = NETIO_PKT_METADATA(packet);
  unsigned char* data = NETIO_PKT_L3_DATA_M(metadata, packet);
  int bytes_in_packet = NETIO_PKT_L3_LENGTH_M(metadata, packet);
  int i;

  unsigned char* iaped_already =
    (unsigned char*)((((unsigned int)metadata) & -64) + 128);
  if (data < iaped_already)
    {
      unsigned int diff = iaped_already - data;
      data += diff;
      bytes_in_packet -= diff;
    }

  for (i = 0; i < bytes_in_packet; i += 16)
    __insn_inv(data + i);

  for (i = 0; i < bytes_in_packet && i < 64*8; i += 64)
    __insn_prefetch(data + i);
}

struct pcap_packet {
  struct pcap_pkthdr      header;
  unsigned int            ordinal;
  const unsigned char *   data;       // L2 data
};


typedef struct _IPHdr
{
  u_int8_t ip_verhl;      /* version & header length */
  u_int8_t ip_tos;        /* type of service */
  u_int16_t ip_len;       /* datagram length */
  u_int16_t ip_id;        /* identification  */
  u_int16_t ip_off;       /* fragment offset */
  u_int8_t ip_ttl;        /* time to live field */
  u_int8_t ip_proto;      /* datagram protocol */
  u_int16_t ip_csum;      /* checksum */
  struct in_addr ip_src;  /* source IP */
  struct in_addr ip_dst;  /* dest IP */
}  __attribute__((__packed__))     IPHdr;

typedef struct _TCPHdr
{
  u_int16_t th_sport;     /* source port */
  u_int16_t th_dport;     /* destination port */
  u_int32_t th_seq;       /* sequence number */
  u_int32_t th_ack;       /* acknowledgement number */
  u_int8_t th_offx2;      /* offset and reserved */
  u_int8_t th_flags;
  u_int16_t th_win;       /* window */
  u_int16_t th_sum;       /* checksum */
  u_int16_t th_urp;       /* urgent pointer */

}    __attribute__((__packed__))      TCPHdr;

#define ETHERNET_HEADER_LEN             14
#define IP_HLEN(iph)   ((iph)->ip_verhl & 0x0f)


#ifdef HASH_PACKETS

static inline unsigned int
tile_hash(const unsigned char *L2_data)
{
  unsigned int src = ntohl(((IPHdr*)(L2_data + ETHERNET_HEADER_LEN))->ip_src.s_addr);
  unsigned int dst = ntohl(((IPHdr*)(L2_data + ETHERNET_HEADER_LEN))->ip_dst.s_addr);
  //unsigned int proto = *((IPHdr*)(L2_data + ETHERNET_HEADER_LEN+1));
  unsigned int proto = ((IPHdr*)(L2_data + ETHERNET_HEADER_LEN))->ip_proto;
  unsigned int hlen = IP_HLEN(((IPHdr*)(L2_data + ETHERNET_HEADER_LEN)));
  uint hash;
  uint sport = 0;
  uint dport = 0;
  hash = __insn_crc32_32(proto, src^dst);
  if (proto == 6)
    {
      sport = ntohs(((TCPHdr*)(L2_data + ETHERNET_HEADER_LEN + hlen*4))->th_sport);
      dport = ntohs(((TCPHdr*)(L2_data + ETHERNET_HEADER_LEN + hlen*4))->th_dport);
      hash = __insn_crc32_32(hash, sport^dport);
    }
  hash = hash ^ (hash>>16);
  uint tile = hash % tile_count;
  //printf("src=%X, dst=%X sport=%d, dport=%d, proto=%d hash=%X tile=%d/%d\n", src, dst,sport, dport, proto, hash, tile, tile_count);

  return tile;
}
#endif


static int
pcap_file_loop(pcap_t *handle, int cnt, pcap_handler callback, u_char *userdata)
{
  struct pcap_pkthdr_with_ordinal pcap_header_with_ordinal;
  struct pcap_pkthdr * pcap_header = &pcap_header_with_ordinal.pkthdr;
  struct pcap_packet *  packet_desc = malloc_shared(MAX_FILE_PACKETS*sizeof(struct pcap_packet));
  int count = 0;
  int my_rank = ilib_group_rank(ILIB_GROUP_SIBLINGS);
  int group_size = ilib_group_size(ILIB_GROUP_SIBLINGS);
  int ordinal = 0;

  //printf("Reading packets into memory\n");

  //
  // Read packets into memory
  //
  for (ordinal = 0, count=0;  fread(&(packet_desc[count].header), sizeof(struct pcap_pkthdr), 1, input_fp) > 0; ordinal++)
    {
      int len = packet_desc[count].header.caplen;
      int buffer_len = (len < 128 ? 128 : (len < 2048 ? 2048 : 64*1024));
      packet_desc[count].data = memalign(64, buffer_len);
      packet_desc[count].ordinal = ordinal;

      fread((void*)packet_desc[count].data, 1, len, input_fp);
#ifdef HASH_PACKETS
      // Hashing
      if (tile_count == 1 || tile_hash(packet_desc[count].data) == my_rank)
#else
	// Round Robin
	if ( (ordinal % group_size) == my_rank)
#endif
	  { // Only hash if there is more than one tile
	    count++;
	    if (count > MAX_FILE_PACKETS)
	      {
		ilib_die("Exceeded the maximum number of packets when reading from a file (%d)\n", MAX_FILE_PACKETS);
	      }
	  }
	else
	  { // Don't keep the packet
	    free((void*)(packet_desc[count].data));
	  }
    }
  fclose(input_fp);
  //printf("Tile:%d Read %d packets from tcpdump file\n", my_rank, count);

#ifdef PROFILE_DETECT
  profiler_clear();
  profiler_chip_clear(SIM_CHIP_ALL);
#endif

  ilib_msg_barrier(ILIB_GROUP_SIBLINGS);

#ifdef PROFILE_DETECT
  profiler_enable();
#endif


#ifdef MEASURE_PERFORMANCE
  pcap_start_cycle = get_cycle_count();
#endif

  //
  // Send the packets into Snort
  //
  for (int p = 0; p < count; p++)
    {
      int packet_len = packet_desc[p].header.len;
      int caplen = packet_desc[p].header.caplen;
      if (caplen > handle->snapshot)
	caplen = handle->snapshot;

      handle->md.stat.ps_recv++;

      pcap_header->caplen	= caplen;
      pcap_header->len	= packet_len;
      pcap_header_with_ordinal.ordinal = packet_desc[p].ordinal;
      gettimeofday(&pcap_header->ts, NULL);


      //BEGIN_Packet();
      callback(userdata, pcap_header, packet_desc[p].data);
      //END_Packet();
    }

#ifdef MEASURE_PERFORMANCE
  pcap_end_cycle = get_cycle_count();
#endif

#ifdef PROFILE_DETECT
  profiler_disable();
#endif


  return count;
}


static int pcap_wait_for_packet(netio_pkt_t *packet, int cnt, unsigned int *packets_to_report)
{
  int no_packet_count = 0;

#ifdef PROFILE_NETIO
  uint64_t start_cycle = get_cycle_count();
  int failed_requests = 1;
#endif

  // Check for simulation termination condition

  while (netio_get_packet(&ipp_queue, packet) != NETIO_PKT)
    {
#ifdef PROFILE_NETIO
      failed_requests++;
#endif
      if (no_packet_count++ > 500)
	{
	  //
	  // Only update the total counter after we have to recevied any packets for
	  // some period of time.
	  //
	  if (*packets_to_report > 0)
	    {
	      ilib_mutex_lock(&	((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex)->lock);
	      ((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex)->total_packets += *packets_to_report;
	      ilib_mutex_unlock(&	((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex)->lock);
	      *packets_to_report = 0;

	    }
	  //PASS(	((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex)->total_packets);
	  if (	((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex)->total_packets >= cnt && cnt >= 0)
	    {
	      return 1;
	    }
	  no_packet_count = 0;
	}
    }

#ifdef PROFILE_NETIO
  cycles_waiting_for_packets += get_cycle_count() - start_cycle;
  empty_requests += failed_requests;
#endif

  return 0;
}

static int
pcap_live_loop(pcap_t *handle, int cnt, pcap_handler callback, u_char *userdata)
{
  int packets = 0, debug = 0;

  struct pcap_pkthdr_with_ordinal	pcap_header_with_ordinal;
  struct pcap_pkthdr *	pcap_header = &(pcap_header_with_ordinal.pkthdr);
  unsigned int packets_to_report = 0;

  if (!ipp_initialized)
    {
      ipp_initialized = 1;
      //
      // Configure IPP
      //
      //
      // Only called from one tile
      configure_ipp(&ipp_queue);

      if(debug) printf("Started pcap_live_loop() on (%d, %d)\n", spr_x_coordinate(), spr_y_coordinate());
      fflush(stdout);
    }

  if(debug) printf("pcap_live_loop(handle=%p)\n", handle);

  if(debug) printf("[2] total_packets_mutex=%p [rank=%d]\n",
		   ((TotalPacketsMutex*)readWriteGlobals->total_packets_mutex),
		   ilib_group_rank(ILIB_GROUP_SIBLINGS));

#ifdef PROFILE_DETECT
  profiler_disable();
  profiler_clear();
  profiler_chip_clear(SIM_CHIP_ALL);
#endif

  ilib_msg_barrier(ILIB_GROUP_SIBLINGS);

#ifdef PROFILE_DETECT
  profiler_enable();
#endif


  while (1)
    {
      netio_pkt_t packet;

      if (netio_get_packet(&ipp_queue, &packet) != NETIO_PKT)
	{

#ifdef MEASURE_PERFORMANCE
	  pcap_end_cycle = get_cycle_count();
#endif
	  if (pcap_wait_for_packet(&packet, cnt, &packets_to_report))
	    break;
	}
#ifdef MEASURE_PERFORMANCE
      if (pcap_start_cycle == 0)
	{
	  pcap_start_cycle = get_cycle_count();
	}
#endif

      //uint hash                = NETIO_PKT_FLOW_HASH(&packet);
      //uint bucket              = NETIO_PKT_BUCKET(&packet);
      uint packet_len          = NETIO_PKT_L2_LENGTH(&packet);
      unsigned char* data      = NETIO_PKT_L2_DATA(&packet);
      unsigned int   ordinal   = NETIO_PKT_ORDINAL(&packet);

      //--// Prefetch packet header
      //--__insn_prefetch(data);
      //--__insn_prefetch(data+64);

      int good_packet = 1;

      if (NETIO_PKT_L3_CSUM_CALCULATED(&packet))
	{
	  good_packet = NETIO_PKT_L3_CSUM_CORRECT(&packet);
	}
#if 0
      if (data[9] == 6)
	{  // TCP packet
	  good_packet &= NETIO_PKT_GOOD_CHECKSUM(&packet);
	}
#endif
      if (good_packet)
	{

	  handle->md.stat.ps_recv++;
	  packets++;
	  packets_to_report++;  // part of total not yet added to global counter

#ifdef PROFILE_NETIO
	  packets_processed++;
#endif

	  //uint packet_number       = NETIO_PKT_ORDINAL(&packet);
	  //printf("Snort:%d - Packet#%d, len=%d, layer3 len=%d\n", my_rank, packet_number, packet_len,NETIO_PKT_L3_LENGTH(&packet) );

	  int caplen = packet_len;
	  if (caplen > handle->snapshot)
	    caplen = handle->snapshot;

	  gettimeofday(&pcap_header->ts, NULL);
	  pcap_header->caplen   = caplen;
	  pcap_header->len	   = packet_len;
	  pcap_header_with_ordinal.ordinal  = ordinal;

	  //PASS(packet_number);
	  callback(userdata, pcap_header, data);
	}
      else
	{
	  // Bad Checksum, drop packet
	  handle->md.stat.ps_drop++;
	  PASS(666);
	}

      invalidate_packet(data, packet_len);
      netio_free_buffer(&ipp_queue, &packet);
    }

#ifdef PROFILE_DETECT
  profiler_disable();
#endif

#ifdef PROFILE_NETIO
  printf("%2d: (%d,%d) non-fulfilled netio requests = %7d, cycles waiting: %9llu, %6d packets processed\n",
         ilib_group_rank(ILIB_GROUP_SIBLINGS), spr_x_coordinate(), spr_y_coordinate(), empty_requests, cycles_waiting_for_packets, packets_processed);
#endif

  return packets;
}

int
pcap_loop(pcap_t *handle, int cnt, pcap_handler callback, u_char *userdata)
{
  if (input_fp)
    {
      return pcap_file_loop(handle, cnt, callback, userdata);
    }
  else
    {
      return pcap_live_loop(handle, cnt, callback, userdata);
    }
}

static void
pcap_oneshot(u_char *userData, const struct pcap_pkthdr *h, const u_char *pkt)
{
  struct singleton *sp = (struct singleton *)userData;
  *sp->hdr = *h;
  sp->pkt = pkt;
}

const u_char *
pcap_next(pcap_t *p, struct pcap_pkthdr *h)
{
  struct singleton s;

  s.hdr = h;
  if (pcap_dispatch(p, 1, pcap_oneshot, (u_char*)&s) <= 0)
    return (0);
  return (s.pkt);
}

int pcap_stats(pcap_t *pd, struct pcap_stat *ps)
{
  memcpy(ps, &(pd->md.stat), sizeof(struct pcap_stat));
  return 0;
}

int	pcap_setfilter(pcap_t *a, struct bpf_program *b)
{
  //FAIL(2);

  return 0;
}

/*
 * Get the current non-blocking mode setting, under the assumption that
 * it's just the standard POSIX non-blocking flag.
 *
 * We don't look at "p->nonblock", in case somebody tweaked the FD
 * directly.
 */
#ifndef WIN32
int
pcap_getnonblock_fd(pcap_t *p, char *errbuf)
{
  return 1; // KMS
}
#endif

int
pcap_setnonblock(pcap_t *p, int nonblock, char *errbuf)
{
  return p->setnonblock_op(p, nonblock, errbuf);
}


void
pcap_perror(pcap_t *p, char *prefix)
{
  fprintf(stderr, "%s: %s\n", prefix, p->errbuf);
}

/*
 * Not all systems have strerror().
 */
char *
pcap_strerror(int errnum)
{
#ifdef HAVE_STRERROR
  return (strerror(errnum));
#else
  extern int sys_nerr;
  extern const char *const sys_errlist[];
  static char ebuf[20];

  if ((unsigned int)errnum < sys_nerr)
    return ((char *)sys_errlist[errnum]);
  (void)snprintf(ebuf, sizeof ebuf, "Unknown error: %d", errnum);
  return(ebuf);
#endif
}

char *
pcap_geterr(pcap_t *p)
{
  return (p->errbuf);
}

int	pcap_compile(pcap_t *a, struct bpf_program *b, char *c, int d,
		     bpf_u_int32 e)
{
  //FAIL(3);
  return 0;
}

int	pcap_compile_nopcap(int a, int b, struct bpf_program * c,
			    char * d, int e, bpf_u_int32 f)
{
  //FAIL(4);
  return 0;
}

void	pcap_freecode(struct bpf_program *a)
{
  //FAIL(5);
}


int
pcap_datalink(pcap_t *p)
{
  return (p->linktype);
}

int
pcap_snapshot(pcap_t *p)
{
  return (p->snapshot);
}

int
pcap_is_swapped(pcap_t *p)
{
  return (p->sf.swapped);
}

int
pcap_major_version(pcap_t *p)
{
  return (p->sf.version_major);
}

int
pcap_minor_version(pcap_t *p)
{
  return (p->sf.version_minor);
}




