/*
 *  Copyright (C) 2007 Luca Deri <deri@ntop.org>
 *  Copyright 2007 Tilera Corporation. All Rights Reserved.
 *
 *  			 http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* ************************************************ */

void tilera_init() {
  size_t init_size;
  ThreadInit *thread;

  profiler_disable();
  profiler_clear();
  profiler_chip_clear(SIM_CHIP_ALL);

  ilib_init();

  /*
    At this point, we could be running the initial program startup
    or reentering to start a new thread (if the iLib init block is
    not NULL).
  */

  thread = ilib_proc_get_init_block(&init_size);

  if(thread) {
    /*
      We where called from ilib_proc_exec(), so spawn a new thread
      as this is basically a thread started by the main program
    */
    readWriteGlobals = thread->shared_vars;
    memcpy(&readOnlyGlobals, &thread->shared_read_only, sizeof(readOnlyGlobals));
    thread->thread_func(thread->argument); /* Run the thread */

    ilib_finish();
    exit(0);
  } else {
    /* We're the main thread so continue the initialization */
  }
}

/* ************************************************ */

void tilera_term() {
  ilib_finish();
}
